"""
    This module contains the class definitions for all things
    regarding the guild and also defines a USER class from the
    _BaseGUILD class.
"""
from    __future__ import annotations
import asyncio
from    contextlib import suppress
from    typing import Any, Literal, Union, List, Optional

from framework import misc
from    .exceptions import *
from    .tracing import *
from    .const import *
from    .message import *
from    . import client
from    . import sql
from    . import core
import  _discord as discord
import  time
import  json
import  pathlib
import  shutil

__all__ = (
    "GUILD",
    "USER"
)

#######################################################################
# Globals
#######################################################################
class GLOBALS:
    """
    Contains the global variables for the module.
    """
    server_log_path = None


class _BaseGUILD:
    """ 
    Represents an universal guild.
    
    .. versionchanged:: 
        v2.0

            - Added the update method.
            - snowflake parameter can now be a discord.Object object.


    Parameters
    ---------------
    snowflake: Union[int, discord.Object]
        Discord's snowflake id or a Discord object that has the ID attribute.
    generate_log: bool
        Set to True if you wish to have message logs for this guild.
    """

    __slots__ = (       # Faster attribute access
        "apiobject",
        "logging",
        "_messages",
    )
    __logname__ = "_BaseGUILD" # Dummy to demonstrate correct definition for @sql._register_type decorator

    def __init__(self,
                 snowflake: Union[int, discord.Object],
                 messages: Optional[List]=[],
                 logging: Optional[bool]=False) -> None:

        self.apiobject: discord.Object = snowflake
        self.logging: bool= logging
        self._messages: list = messages  # Contains all the different message objects, this gets sorted in `.initialize()` method

    @property
    def _log_file_name(self):
        """
        Returns a string that transforms the xGUILD's discord name into
        a string that contains only allowed character. This is a method instead of
        property because the name can change overtime.
        """
        raise NotImplementedError
    
    @property
    def messages(self) -> List[BaseMESSAGE]:
        """
        Returns all the (initialized) message objects inside the object.

        .. versionadded:: v2.0
        """
        ret = []
        for x in self.__slots__:
            if hasattr(self, x) and not x.startswith("_") and x.endswith("messages"):
                ret.extend(getattr(self , x))
        return ret

    @property
    def snowflake(self) -> int:
        """
        .. versionadded:: v2.0

        Returns the discord's snowflake ID.
        """
        return self.apiobject if isinstance(self.apiobject, int) else self.apiobject.id

    def __eq__(self, other: _BaseGUILD) -> bool:
        """
        Compares two guild objects if they're equal.
        """
        return self.snowflake == other.snowflake

    async def add_message(self, message: BaseMESSAGE):
        """
        Adds a message to the message list.

        Parameters
        -----------
        message: BaseMESSAGE
            Message object to add.
        """
        raise NotImplementedError

    async def initialize(self):
        """
        Initializes the guild and then any message objects
        the guild may hold.
        """
        raise NotImplementedError

    async def advertise(self,
                        mode: Literal["text", "voice"]):
        """
        Main coroutine responsible for sending all the messages to this specific guild,
        it is called from the core module's advertiser task.
        
        Parameters
        --------------
        mode: Literal["text", "voice"]
            Tells which task called this method (there is one task for textual messages and one for voice like messages).
        """
        raise NotImplementedError

    async def update(self, **kwargs):
        """
        .. versionadded:: v2.0

        Used for changing the initialization parameters the object was initialized with.
        
        .. warning:: 
            This method will BLOCK until every message has finished shilling!
            This is done for safety due to asynchronous operations.

        .. warning::
            Upon updating, the internal state of objects get's reset, meaning you basically have a brand new created object.
        
        Parameters
        -------------
        **kwargs: Any
            Custom number of keyword parameters which you want to update, these can be anything that is available during the object creation.
        """
        raise NotImplementedError

    async def generate_log(self,
                           message_context: dict) -> None:
        """
        Generates a .json type log or logs to a SQL database for each message send attempt.

        Prioritizes SQL logging.

        Parameter:
        -----------
        data_context: dict
            Dictionary containing data describing the message send attempt. (Return of ``message.send()``)
        """
        guild_context = {
            "name" : str(self.apiobject),
            "id" : self.snowflake,
            "type" : type(self).__logname__,
        }

        try:
            # Try to obtain the sql manager. If it returns None (sql logging disabled), save into file
            manager = sql.get_sql_manager()
            if (
                manager is None or  # Short circuit evaluation
                not await manager._save_log(guild_context, message_context)
            ):
                timestruct = time.localtime()
                timestamp = "{:02d}.{:02d}.{:04d} {:02d}:{:02d}:{:02d}".format(timestruct.tm_mday,
                                                                            timestruct.tm_mon,
                                                                            timestruct.tm_year,
                                                                            timestruct.tm_hour,
                                                                            timestruct.tm_min,
                                                                            timestruct.tm_sec)
                logging_output = pathlib.Path(GLOBALS.server_log_path)\
                            .joinpath("{:02d}".format(timestruct.tm_year))\
                            .joinpath("{:02d}".format(timestruct.tm_mon))\
                            .joinpath("{:02d}".format(timestruct.tm_mday))
                with suppress(FileExistsError):
                    logging_output.mkdir(parents=True,exist_ok=True)

                logging_output = str(logging_output.joinpath(self._log_file_name))

                # Create file if it doesn't exist
                fresh_file = False
                with suppress(FileExistsError), open(logging_output, "x", encoding="utf-8"):
                    fresh_file = True

                # Write to file
                with open(logging_output,'r+', encoding='utf-8') as appender:
                    appender_data = None
                    appender.seek(0) # Append moves cursor to the end of the file
                    try:
                        appender_data = json.load(appender)
                    except json.JSONDecodeError:
                        # No valid json in the file, create new data
                        # and create a .old file to store this invalid data
                        # Copy-paste to .old file to prevent data loss
                        if not fresh_file: ## Failed because the file was just created, no need to copy-paste
                            # File not newly created, and has invalid data
                            shutil.copyfile(logging_output, f"{logging_output}.old")
                        # Create new data
                        appender_data = {}
                        appender_data["name"] = guild_context["name"]
                        appender_data["id"]   = guild_context["id"]
                        appender_data["type"] = guild_context["type"]
                        appender_data["message_history"] = []
                    finally:
                        appender.seek(0) # Reset cursor to the beginning of the file after reading

                    appender_data["message_history"].insert(0,
                        {
                            **message_context,
                            "index":    appender_data["message_history"][0]["index"] + 1 if len(appender_data["message_history"]) else 0,
                            "timestamp": timestamp
                        })                
                    json.dump(appender_data, appender, indent=4)
                    appender.truncate() # Remove any old data

        except Exception as exception:
            # Any uncaught exception (prevent from complete framework stop)
            trace(f"[{type(self).__name__}]: Unable to save log. Exception: {exception}", TraceLEVELS.WARNING)


@sql._register_type("GuildTYPE")
class GUILD(_BaseGUILD):
    """
    The GUILD object represents a server to which messages will be sent.
    
    .. versionchanged:: v2.0

        - Added the update method
        - Renamed param ``guild_id`` to ``snowflake``
        - Renamed ``generate_log`` parameter to ``logging``
        - snowflake parameter can now be a discord.Guild object.

    Parameters
    ------------
    snowflake: Union[int, discord.Guild]
        Discord's snowflake ID of the guild or discord.Guild object.
    messages: Optional[List[Union[TextMESSAGE, VoiceMESSAGE]]]
        Optional list of TextMESSAGE/VoiceMESSAGE objects.
    logging:  Optional[bool]
        Optional variable dictating whatever to log sent messages inside this guild.
    """
    
    __logname__ = "GUILD" # For sql._register_type
    __slots__   = (
        "t_messages",
        "vc_messages",
        "update_semaphore",
    )

    def __init__(self,
                 snowflake: Union[int, discord.Guild],
                 messages: Optional[List[Union[TextMESSAGE, VoiceMESSAGE]]]=[],
                 logging: Optional[bool]=False):
        super().__init__(snowflake, messages, logging)
        self.t_messages: List[TextMESSAGE] = []
        self.vc_messages: List[VoiceMESSAGE] = []
        misc._write_attr_once(self, "update_semaphore", asyncio.Semaphore(2))
    
    @property
    def _log_file_name(self):
        return "".join(char if char not in C_FILE_NAME_FORBIDDEN_CHAR else "#" for char in self.apiobject.name) + ".json"
    
    async def add_message(self, message: Union[TextMESSAGE, VoiceMESSAGE]):
        """
        Adds a message to the message list.

        Parameters
        --------------
        message: Union[TextMESSAGE, VoiceMESSAGE]
            Message object to add.

        Raises
        --------------
        DAFParameterError(code=DAF_INVALID_TYPE)
            Raised when the message is not of type TextMESSAGE or VoiceMESSAGE.
        Other
            Raised from message.initialize() method.
        """
        if not isinstance(message, (TextMESSAGE, VoiceMESSAGE)):
            raise DAFParameterError(f"Invalid xxxMESSAGE type: {type(message).__name__}, expected  {TextMESSAGE.__name__} or {VoiceMESSAGE.__name__}", DAF_INVALID_TYPE)

        await message.initialize()

        if isinstance(message, TextMESSAGE):
            self.t_messages.append(message)
        elif isinstance(message, VoiceMESSAGE):
            self.vc_messages.append(message)

    def remove_message(self, message: Union[TextMESSAGE, VoiceMESSAGE]):
        """
        Removes a message from the message list.

        Parameters
        --------------
        message: Union[TextMESSAGE, VoiceMESSAGE]
            Message object to remove.

        Raises
        --------------
        DAFParameterError(code=DAF_INVALID_TYPE)
            Raised when the message is not of type TextMESSAGE or VoiceMESSAGE.
        ValueError
            Raised when the message is not present in the list.
        """
        if isinstance(message, TextMESSAGE):
            message._delete()
            self.t_messages.remove(message)
            return
        elif isinstance(message, VoiceMESSAGE):
            message._delete()
            self.vc_messages.remove(message)
            return

        raise DAFParameterError(f"Invalid xxxMESSAGE type: {type(message).__name__}, expected  {TextMESSAGE.__name__} or {VoiceMESSAGE.__name__}", DAF_INVALID_TYPE)

    async def initialize(self) -> None:
        """
        This function initializes the API related objects and then tries to initialize the MESSAGE objects.

        Raises
        -----------
        DAFNotFoundError(code=DAF_GUILD_ID_NOT_FOUND)
            Raised when the guild_id wasn't found.

        Other
            Raised from .add_message(message_object) method.
        """
        guild_id = self.snowflake
        if isinstance(self.apiobject, int):
            cl = client.get_client()
            self.apiobject = cl.get_guild(guild_id)

        if self.apiobject is not None:
            for message in self._messages:
                await self.add_message(message)

            self._messages.clear()
            return

        raise DAFNotFoundError(f"Unable to find guild with ID: {guild_id}", DAF_GUILD_ID_NOT_FOUND)

    @misc._async_safe("update_semaphore", 1)
    async def advertise(self,
                        mode: Literal["text", "voice"]):
        """
        Main coroutine responsible for sending all the messages to this specific guild,
        it is called from the core module's advertiser task.
        
        Parameters
        --------------
        mode: Literal["text", "voice"]
            Tells which task called this method (there is one task for textual messages and one for voice like messages).
        """
        msg_list = self.t_messages if mode == "text" else self.vc_messages
        for message in msg_list[:]: # Copy the avoid issues with the list being modified while iterating (add_message/remove_message)
            if message.is_ready():
                message.reset_timer()
                message_ret = await message.send()
                # Check if the message still has any channels (as they can be auto removed on 404 status)
                if not len(message.channels) and message in msg_list:
                    self.remove_message(message) # All channels were removed (either not found or forbidden) -> remove message from send list
                    trace(f"[GUILD]: Removing a {type(message).__name__} because it's channels were removed, in guild {self.apiobject.name}(ID: {self.snowflake})", TraceLEVELS.WARNING)
                if self.logging and message_ret is not None:
                    await self.generate_log(message_ret)              
            

    @misc._async_safe("update_semaphore", 2) # Take 2 since 2 tasks share access
    async def update(self, **kwargs):
        """
        Used for changing the initialization parameters the object was initialized with.
        
        .. versionadded:: v2.0

        .. warning:: 
            Upon updating, the internal state of objects get's reset, meaning you basically have a brand new created object.
            It also resets the message objects.
        
        Parameters
        -------------
        **kwargs: Any
            Custom number of keyword parameters which you want to update, these can be anything that is available during the object creation.

        Raises
        -----------
        DAFParameterError(code=DAF_UPDATE_PARAMETER_ERROR)
            Invalid keyword argument was passed.
        Other
            Raised from .initialize() method.
        """
        # Update the guild
        if "snowflake" not in kwargs:
            kwargs["snowflake"] = self.snowflake

        await core._update(self, **kwargs)
        # Update messages
        for message in self.messages:
            await message.update()
        

@sql._register_type("GuildTYPE")
class USER(_BaseGUILD):
    """
    The USER object represents a user to whom messages will be sent.
    
    .. versionchanged:: v2.0

        - ``user_id`` parameter renamed to ``snowflake``
        - ``snowflake`` can now also be a ``discord.User`` object
        - renamed ``generate_log`` parameter to ``logging``.

    Parameters
    ------------
    snowflake: Union[int, discord.User]
        Discord's snowflake ID of the user or discord.User object.
    messages: Optional[List[DirectMESSAGE]]
        Optional list of DirectMESSAGE objects.
    logging: Optional[bool]
        Optional variable dictating whatever to log sent messages inside this guild.
    """

    __logname__ = "USER" # For sql._register_type
    __slots__   = (
        "t_messages",
        "update_semaphore",
    )

    @property
    def _log_file_name(self):
        return "".join(char if char not in C_FILE_NAME_FORBIDDEN_CHAR else "#" for char in f"{self.apiobject.display_name}#{self.apiobject.discriminator}") + ".json"

    def __init__(self,
                 snowflake: Union[int, discord.User],
                 messages: Optional[List[DirectMESSAGE]]=[],
                 logging: Optional[bool] = False) -> None:
        super().__init__(snowflake, messages, logging)
        self.t_messages: List[DirectMESSAGE] = []
        
        misc._write_attr_once(self, "update_semaphore", asyncio.Semaphore(2)) # Only allows re-referencing this attribute once

    async def add_message(self, message):
        """
        Adds a message to the message list.

        Parameters
        --------------
        message: DirectMESSAGE
            Message object to add.

        Raises
        --------------
        DAFParameterError(code=DAF_INVALID_TYPE)
            Raised when the message is not of type DirectMESSAGE.
        Other
            Raised from message.initialize() method.
        """
        if not isinstance(message, DirectMESSAGE):
            raise DAFParameterError(f"Invalid xxxMESSAGE type: {type(message).__name__}, expected  {DirectMESSAGE.__name__}", DAF_INVALID_TYPE)

        await message.initialize(user=self.apiobject)
        self.t_messages.append(message)

    
    def remove_message(self, message: DirectMESSAGE):
        """
        .. versionadded:: v2.0

        Removes a message from the message list.

        Parameters
        --------------
        message: DirectMESSAGE
            Message object to remove.

        Raises
        --------------
        DAFParameterError(code=DAF_INVALID_TYPE)
            Raised when the message is not of type DirectMESSAGE.
        """
        if isinstance(message, DirectMESSAGE):
            message._delete()
            self.t_messages.remove(message)
            return

        raise DAFParameterError(f"Invalid xxxMESSAGE type: {type(message).__name__}, expected  {DirectMESSAGE.__name__}", DAF_INVALID_TYPE)

    async def initialize(self):
        """
        This function initializes the API related objects and then tries to initialize the MESSAGE objects.

        Raises
        -----------
        DAFNotFoundError(code=DAF_USER_CREATE_DM)
            Raised when the DM could not be created.
        Other
            Raised from .add_message(message_object) method.
        """
        user_id = self.snowflake
        if isinstance(self.apiobject, int):
            cl = client.get_client()
            self.apiobject = await cl.get_or_fetch_user(user_id) # Get object from cache

        # Api object was found in cache or fetched from API -> initialize messages
        if self.apiobject is not None:
            for message in self._messages:
                await self.add_message(message)

            self._messages.clear()
            return

        # Api object wasn't found, even after direct API call to discord.
        raise DAFNotFoundError(f"[USER]: Unable to create DM with user id: {user_id}", DAF_USER_CREATE_DM)

    @misc._async_safe("update_semaphore", 1)
    async def advertise(self,
                        mode: Literal["text", "voice"]) -> None:
        """
        Main coroutine responsible for sending all the messages to this specific guild,
        it is called from the core module's advertiser task.
        
        Parameters
        --------------
        mode: Literal["text", "voice"]
            Tells which task called this method (there is one task for textual messages and one for voice like messages).
        """
        if mode == "text":  # Does not have voice messages, only text based (DirectMESSAGE)
            for message in self.t_messages[:]: # Copy the avoid issues with the list being modified while iterating (add_message/remove_message)
                if message.is_ready():
                    message.reset_timer()
                    message_ret = await message.send()
                    if message_ret is not None:
                        if self.logging:
                            await self.generate_log(message_ret)

                        if message.deleted:  # Only True on critical errors that need to be handled by removing all messages
                            for msg in self.t_messages:
                                self.remove_message(msg)

                            trace(f"Removing all messages for user {self.apiobject.display_name}#{self.apiobject.discriminator} (ID: {self.snowflake}) because we do not have permissions to send to that user.", TraceLEVELS.WARNING)
                            break
    
    @misc._async_safe("update_semaphore", 2)
    async def update(self, **kwargs):
        """
        .. versionadded:: v2.0

        Used for changing the initialization parameters the object was initialized with.

        .. warning:: 
            Upon updating, the internal state of objects get's reset, meaning you basically have a brand new created object.
            It also resets the message objects.

        Parameters
        -------------
        **kwargs: Any
            Custom number of keyword parameters which you want to update, these can be anything that is available during the object creation.

        Raises
        -----------
        DAFParameterError(code=DAF_UPDATE_PARAMETER_ERROR)
            Invalid keyword argument was passed.
        Other
            Raised from .initialize() method.
        """
        # Update the guild
        if "snowflake" not in kwargs:
            kwargs["snowflake"] = self.snowflake

        await core._update(self, **kwargs)
        # Update messages
        for message in self.messages:
            await message.update(init_options={"user" : self.apiobject})


"""
    Contains base definitions for different message classes."""

from    typing import Dict, Set, Tuple, Union
from    ..dtypes import *
from    ..tracing import *
from    ..timing import *
from    ..exceptions import *
from    .. import misc
import  random
import  _discord as discord
import  asyncio


__all__ = (
    "BaseMESSAGE",
)

class BaseMESSAGE:
    """
    This is the base class for all the different classes that
    represent a message you want to be sent into discord.
    
    Parameters
    -----------------
    start_period: Union[int, None]
        If this this is not None, then it dictates the bottom limit for range of the randomized period. Set this to None
                                         for a fixed sending period.
    end_period: int
        If start_period is not None, this dictates the upper limit for range of the randomized period. If start_period is None, then this
                            dictates a fixed sending period in SECONDS, eg. if you pass the value `5`, that means the message will be sent every 5 seconds.
    data: inherited class dependant
        The data that will be sent to Discord. Valid data types are defined inside `__valid_data_types__` set.
    start_now: bool
        Dictates if the message should be send immediately after framework start, or wait the period first.
    """
    __slots__ = (
        "randomized_time",
        "period",
        "start_period",
        "end_period",
        "timer",
        "force_retry",
        "data",
        "update_semaphore",
        "_deleted"
    )

    __logname__: str = "" # Used for registering SQL types and to get the message type for saving the log

    # The "__valid_data_types__" should be implemented in the INHERITED classes.
    # The set contains all the data types that the class is allowed to accept, this variable
    # is then checked for allowed data types in the "initialize" function bellow.
    __valid_data_types__ = {}

    def __init__(self,
                start_period : Union[int, None],
                end_period : int,
                data,
                start_now):
        # If start_period is none -> period will not be randomized
        self.start_period = start_period
        self.end_period   = end_period
        if start_period is None:            
            self.randomized_time = False
            self.period = end_period
        else:
            self.randomized_time = True
            self.period = random.randrange(self.start_period, self.end_period)

        self.timer = TIMER()
        self.force_retry = {"ENABLED" : start_now, "TIME" : 0}
        self.data = data
        misc._write_attr_once(self, "_deleted", False)
        misc._write_attr_once(self, "update_semaphore", asyncio.Semaphore(1))

    @property
    def deleted(self) -> bool:
        """
        Property that indicates if an object has been deleted from the shilling list.

        If this is True, you should dereference this object from any variables.
        """
        return self._deleted

    def _delete(self):
        """
        Sets the deleted flag to True, indicating the user should stop
        using this message.
        """
        self._deleted = True

    def _generate_exception(self, 
                           status: int,
                           code: int,
                           description: str,
                           cls: discord.HTTPException) -> discord.HTTPException:
        """
        Generates a discord.HTTPException inherited class exception object.
        This is used for generating dummy exceptions that are then raised inside the `._send_channel()`
        method to simulate what would be the result of a API call, without actually having to call the API (reduces the number of bad responses).

        Parameters
        -------------
        status: int
            Discord status code of the exception.
        code: int
            Discord error code.
        description: str
            The textual description of the error.
        cls: discord.HTTPException
            Inherited class from discord.HTTPException to make exception from.
        """
        resp = Exception()
        resp.status = status
        resp.status_code = status
        resp.reason = cls.__name__
        resp = cls(resp, {"message" : description, "code" : code})
        return resp

    def _generate_log_context(self):
        """
        This method is used for generating a dictionary (later converted to json) of the
        data that is to be included in the message log. This is to be implemented inside the
        inherited classes.
        """
        raise NotImplementedError
    
    def _get_data(self) -> dict:
        """
        Returns a dictionary of keyword arguments that is then expanded
        into other functions (_send_channel, generate_log)
        This is to be implemented in inherited classes due to different data_types
        """
        raise NotImplementedError

    def is_ready(self) -> bool:
        """ 
        This method returns bool indicating if message is ready to be sent
        """
        return (not self.force_retry["ENABLED"] and self.timer.elapsed() > self.period or
                self.force_retry["ENABLED"] and self.timer.elapsed() > self.force_retry["TIME"])

    def reset_timer(self) -> None:
        """ 
        Resets internal timer (and force period)
        """
        self.timer.reset()
        self.timer.start()
        self.force_retry["ENABLED"] = False
        if self.randomized_time is True:
            self.period = random.randrange(self.start_period, self.end_period)

    async def _send_channel(self) -> dict:
        """
        Sends data to a specific channel, this is separate from send
        for easier implementation of similar inherited classes
        The method returns a dictionary: `{"success": bool, "reason": discord.HTTPException}` where 
        `"reason"` is only present if `"success"` `is False`
        """
        raise NotImplementedError

    async def send(self) -> dict:
        """
        Sends a message to all the channels.
        Returns a dictionary generated by the `._generate_log_context` method
        """
        raise NotImplementedError

    async def _initialize_channels(self):
        """
        This method initializes the implementation specific
        api objects and checks for the correct channel input context.
        """
        raise NotImplementedError

    async def initialize_data(self):
        """
        This method checks for the correct data input to the xxxMESSAGE
        object. The expected datatypes for specific implementation is
        defined thru the static variable __valid_data_types__

        Raises
        ------------
        - `DAFParameterError(code=DAF_INVALID_TYPE)` - Raised when a parameter is of invalid type
        - `DAFNotFoundError(code=DAF_MISSING_PARAMETER)` - Raised when no data parameters were passed.
        """

        # Check for correct data types of the MESSAGE.data parameter
        if not isinstance(self.data, _FunctionBaseCLASS):
            # This is meant only as a pre-check if the parameters are correct so you wouldn't eg. start
            # sending this message 6 hours later and only then realize the parameters were incorrect.
            # The parameters also get checked/parsed each period right before the send.

            # Convert any arguments passed into a list of arguments
            if isinstance(self.data, (list, tuple, set)):
                self.data = list(self.data)   # Convert into a regular list to allow removal of items
            else:
                self.data = [self.data]       # Place into a list for iteration, to avoid additional code

            # Check all the arguments
            for data in self.data[:]:
                # Check all the data types of all the passed to the data parameter.
                # If class does not match the allowed types, then the object is removed.
                # The for loop iterates thru a shallow copy (sliced list) of data_params to allow removal of items
                # without affecting the iteration (would skip elements without a copy or use of while loop).

                # The inherited classes MUST DEFINE THE "__valid_data_types__" inside the class which should be a set of the allowed data types

                if (
                        type(data) not in type(self).__valid_data_types__
                    ):
                    if isinstance(data, _FunctionBaseCLASS):
                        raise DAFParameterError(f"The function can only be used on the data parameter directly, not in a iterable. Function: {data.func_name}", DAF_INVALID_TYPE)
                    else:
                        trace(f"INVALID DATA PARAMETER PASSED!\nArgument is of type : {type(data).__name__}\nSee README.md for allowed data types", TraceLEVELS.WARNING)
                        raise DAFParameterError(f"Invalid data type {type(data).__name__}. Allowed types: {type(self).__valid_data_types__}", DAF_INVALID_TYPE)

            if len(self.data) == 0:
                raise DAFNotFoundError(f"No data parameters were passed", DAF_MISSING_PARAMETER)

    async def update(self, init_options: dict={}, **kwargs):
        """
        Used for changing the initialization parameters the object was initialized with.
        
        .. warning::
            Upon updating, the internal state of objects get's reset, meaning you basically have a brand new created object.     
        
        Parameters
        -------------
        init_options: dict
            Contains the initialization options used in .initialize() method for re-initializing certain objects.
            This is implementation specific and not necessarily available.
        original_params:
            The allowed parameters are the initialization parameters first used on creation of the object AND 
        
        Raises
        ------------
        DAFParameterError(code=DAF_UPDATE_PARAMETER_ERROR)
            Invalid keyword argument was passed
        Other
            Raised from .initialize() method
        
        .. versionadded::
            v2.0
        """

        raise NotImplementedError
        
    async def initialize(self, **options):
        """
        The initialize method initializes the message object.
    
        Parameters
        -------------
        - options - keyword arguments sent to _initialize_channels() from an inherited (from _BaseGUILD) class, contains extra init options.
        
        Raises
        -------------
        - Exceptions raised from ._initialize_channels() and .initialize_data() methods
        """

        await self._initialize_channels(**options)
        await self.initialize_data()




"""
Contains definitions for message classes that are text based (TextMESSAGE & DirectMESSAGE)."""

from   .base        import *
from   ..           import client
from   ..           import sql
from   ..           import misc
from   ..dtypes     import *
from   ..tracing    import *
from   ..const      import *
from   ..exceptions import *
from   ..           import core
from   typing       import Any, Dict, List, Iterable, Set, Union, Literal
import asyncio
import _discord as discord


__all__ = (
    "TextMESSAGE",
    "DirectMESSAGE"
)

@sql._register_type("MessageTYPE")
class TextMESSAGE(BaseMESSAGE):
    """
    This class is used for creating objects that represent messages which will be sent to Discord's TEXT CHANNELS.

    .. versionchanged:: v2.0
            
        - Renamed ``channel_ids`` parameter to ``channels``
        - Channels parameter now also accepts channel objects instead of int.
        - .update method added.

    Parameters
    ------------
    start_period: Union[int, None]
        The value of this parameter can be:

        ..  table:: 
        
            ===========  =================================================================================================================
             Value        Info
            ===========  =================================================================================================================
             None         Messages are sent in a constant time period equal to the value of ``end_period``.
             int > 0      Messages are sent in a randomized time period. ``start_period`` represents the bottom limit of this period.
            ===========  =================================================================================================================

    end_period: int
        If ``start_period`` > 0, then this represents the upper limit of randomized time period in which messages will be sent.
        If ``start_period`` is None, then this represents the actual time period between each message send.

        .. code-block:: python
            :caption: **Randomized** sending period between **5** seconds and **10** seconds.
            
            # Time between each send is somewhere between 5 seconds and 10 seconds.
            framework.TextMESSAGE(start_period=5, end_period=10, data="Second Message", channels=[12345], mode="send", start_now=True)

        .. code-block:: python
            :caption: **Fixed** sending period at **10** seconds

            # Time between each send is exactly 10 seconds.
            framework.TextMESSAGE(start_period=None, end_period=10, data="Second Message", channels=[12345], mode="send", start_now=True)

    data: Union[str, EMBED, FILE, List[Union[str, EMBED, FILE]], _FunctionBaseCLASS]
        The data parameter is the actual data that will be sent using discord's API. The data types of this parameter can be:
            - str (normal text),
            - :ref:`EMBED`,
            - :ref:`FILE`,
            - List/Tuple containing any of the above arguments (There can up to 1 string, up to 1 :ref:`EMBED` and up to 10 :ref:`FILE` objects. If more than 1 string or embeds are sent, the framework will only consider the last found).
            - Function that accepts any amount of parameters and returns any of the above types. To pass a function, YOU MUST USE THE :ref:`data_function` decorator on the function before passing the function to the framework.

    channels: Iterable[Union[int, discord.TextChannel, discord.Thread]]
        Channels that it will be advertised into (Can be snowflake ID or channel objects from PyCord).
    mode: str
        Parameter that defines how message will be sent to a channel.
        It can be:   

        .. table:: 
            :align: left

            =================  =======================================================================================
              Mode               Description
            =================  =======================================================================================
             "send"             each period a new message will be sent,                                               
             "edit"             each period the previously send message will be edited (if it exists)                 
             "clear-send"       previous message will be deleted and a new one sent.                                  
            =================  =======================================================================================

    start_now: bool
        If True, then the framework will send the message as soon as it is run.
    """

    __slots__ = (
        "channels",
        "mode",
        "sent_messages",
    )
    __logname__: str = "TextMESSAGE"               # Used for registering SQL types and to get the message type for saving the log
    __valid_data_types__: set = {str, EMBED, FILE} # Used in initialize_data to check if valid parameters were passed

    def __init__(self, start_period: Union[float, None],
                 end_period: float,
                 data: Union[str, EMBED, FILE, List[Union[str, EMBED, FILE]], _FunctionBaseCLASS],
                 channels: Iterable[Union[int, discord.TextChannel, discord.Thread]],
                 mode: Literal["send", "edit", "clear-send"] = "send",
                 start_now: bool = True):
        super().__init__(start_period, end_period, data, start_now)
        self.mode = mode
        self.channels = list(set(channels)) # Automatically removes duplicates
        self.sent_messages = {} # Dictionary for storing last sent message for each channel

    def _generate_log_context(self,
                             text : str,
                             embed : EMBED,
                             files : List[FILE],
                             succeeded_ch: List[Union[discord.TextChannel, discord.Thread]],
                             failed_ch: List[Dict[Union[discord.TextChannel, discord.Thread], Exception]]):
        """
        Generates information about the message send attempt that is to be saved into a log.

        Parameters
        -----------
        text: str
            The text that was sent.
        embed: EMBED
            The embed that was sent.
        files: List[FILE]
            List of files that were sent.
        succeeded_ch: List[Union[discord.TextChannel, discord.Thread]]
            List of the successfully streamed channels.
        failed_ch: failed_ch: List[Dict[Union[discord.TextChannel, discord.Thread], Exception]]
            List of dictionaries contained the failed channel and the Exception object.
        
        Returns
        ----------
        Dict[str, Any]
            Dictionary containing:

                - sent_data: Dict[str, Any]:
                    - text: str
                        The text that was sent.
                    - embed: Dict[str, Any]
                        The embed that was sent.
                    - files: List[str]
                        List of files that were sent.

                - channels: Dict[str, List]:
                    - successful: List[Dict[str, int]] - List of dictionaries containing name of the channel and snowflake id of the channels.
                    - failed: List[Dict[str, Any]] - List of dictionaries containing name of the channel (str), snowflake id (int) and reason why streaming to channel failed (str).
                
                - type: str - The type of the message, this is always TextMESSAGE.
                - mode: str - The mode used to send the message (send, edit, clear-send).
        """

        succeeded_ch = [{"name": str(channel), "id" : channel.id} for channel in succeeded_ch]
        failed_ch = [{"name": str(entry["channel"]), "id" : entry["channel"].id,
                     "reason": str(entry["reason"])} for entry in failed_ch]

        embed = embed.to_dict() if embed is not None else None

        files = [x.filename for x in files]
        sent_data_context = {}
        if text is not None:
            sent_data_context["text"] = text
        if embed is not None:
            sent_data_context["embed"] = embed
        if len(files):
            sent_data_context["files"] = files
        return {
            "sent_data": {
                **sent_data_context
            },
            "channels": {
                "successful" : succeeded_ch,
                "failed": failed_ch
            },
            "type" : type(self).__name__,
            "mode" : self.mode,
        }
    
    def _get_data(self) -> dict:
        """"
        Returns a dictionary of keyword arguments that is then expanded
        into other methods eg. `_send_channel, generate_log`
        """
        data = self.data.get_data() if isinstance(self.data, _FunctionBaseCLASS) else self.data
        _data_to_send = {"embed": None, "text": None, "files": []}
        if data is not None:
            if not isinstance(data, (list, tuple, set)):
                data = (data,)
            for element in data:
                if isinstance(element, str):
                    _data_to_send["text"] = element
                elif isinstance(element, EMBED):
                    _data_to_send["embed"] = element
                elif isinstance(element, FILE):
                    _data_to_send["files"].append(element)
        return _data_to_send

    async def _initialize_channels(self):
        """
        This method initializes the implementation specific api objects and checks for the correct channel input context.
        
        Raises
        ------------
        - `DAFParameterError(code=DAF_INVALID_TYPE)` - Raised when the object retrieved from channels is not a discord.TextChannel or discord.Thread object.
        - `DAFNotFoundError(code=DAF_MISSING_PARAMETER)` - Raised when no channels could be found were parsed.
        """
        ch_i = 0
        cl = client.get_client()
        while ch_i < len(self.channels):
            channel = self.channels[ch_i]
            if isinstance(channel, discord.abc.GuildChannel):
                channel_id = channel.id
            else:
                channel_id = channel
                channel = self.channels[ch_i] = cl.get_channel(channel_id)

            if channel is None:
                trace(f"Unable to get channel from ID {channel_id}", TraceLEVELS.ERROR)
                self.channels.remove(channel)
            elif type(channel) not in {discord.TextChannel, discord.Thread}:
                raise DAFParameterError(f"TextMESSAGE object received channel type of {type(channel).__name__}, but was expecting discord.TextChannel or discord.Thread", DAF_INVALID_TYPE)
            else:
                ch_i += 1

        if not len(self.channels):
            raise DAFNotFoundError(f"No valid channels were passed to {type(self)} object", DAF_MISSING_PARAMETER)
    
    async def _handle_error(self, channel: Union[discord.TextChannel, discord.Thread], ex: Exception) -> bool:
        """
        This method handles the error that occurred during the execution of the function.
        Returns `True` if error was handled.

        Parameters
        -----------
        channel: Union[discord.TextChannel, discord.Thread]
            The channel where the exception occurred.
        ex: Exception
            The exception that occurred during a send attempt.
        """
        handled = False
        if isinstance(ex, discord.HTTPException):
            if ex.status == 429:  # Rate limit
                retry_after = int(ex.response.headers["Retry-After"]) * RLIM_SAFETY_FACTOR
                if ex.code == 20016:    # Slow Mode
                    self.force_retry["ENABLED"] = True
                    self.force_retry["TIME"] = retry_after
                    trace(f"{channel.name} is in slow mode, retrying in {retry_after} seconds", TraceLEVELS.WARNING)
                handled = True
            elif ex.status == 404:      # Unknown object
                if ex.code == 10008:    # Unknown message
                    self.sent_messages[channel.id] = None
                    handled = True
        return handled

    async def _send_channel(self,
                           channel: Union[discord.TextChannel, discord.Thread],
                           text: str,
                           embed: EMBED,
                           files: List[FILE]) -> dict:
        """
        Sends data to specific channel
        
        Returns a dictionary:
        - "success" - Returns True if successful, else False
        - "reason"  - Only present if "success" is False, contains the Exception returned by the send attempt
        
        Parameters
        -------------
        channel: Union[discord.TextChannel, discord.Thread]
            The channel in which to send the data.
        text: str
            The text to send.
        embed: EMBED
            The embedded frame to send.
        files: List[FILE]
            List of files to send.
        """ 

        ch_perms = channel.permissions_for(channel.guild.get_member(client.get_client().user.id))
        for tries in range(3):  # Maximum 3 tries (if rate limit)
            try:
                if ch_perms.send_messages is False: # Check if we have permissions
                    raise self._generate_exception(403, 50013, "You lack permissions to perform that action", discord.Forbidden)

                # Delete previous message if clear-send mode is chosen and message exists
                if self.mode == "clear-send" and self.sent_messages.get(channel.id, None) is not None:
                    await self.sent_messages[channel.id].delete()
                    self.sent_messages[channel.id] = None

                # Send/Edit message
                if  (self.mode in  {"send" , "clear-send"} or # Mode dictates to send new message or delete previous and then send new message or mode dictates edit but message was  never sent to this channel before
                    self.mode == "edit" and self.sent_messages.get(channel.id, None) is None
                    ):
                    discord_sent_msg = await channel.send(  text,
                                                            embed=embed,
                                                            # Create discord.File objects here so it is catched by the except block and then logged
                                                            files=[discord.File(fwFILE.filename) for fwFILE in files])
                    self.sent_messages[channel.id] = discord_sent_msg
                # Mode is edit and message was already send to this channel
                elif self.mode == "edit":
                    await self.sent_messages[channel.id].edit ( text,
                                                                embed=embed)

                return {"success" : True}

            except Exception as ex:
                if not await self._handle_error(channel, ex):
                    return {"success" : False, "reason" : ex}

    @misc._async_safe("update_semaphore")
    async def send(self) -> Union[dict,  None]:
        """
        Sends the data into the channels.
        
        Returns
        ----------
        Union[Dict, None]
            Returns a dictionary generated by the ``_generate_log_context`` method or the None object if message wasn't ready to be sent (:ref:`data_function` returned None or an invalid type)
            
            This is then passed to :ref:`GUILD`.generate_log method.
        """       
        # Acquire mutex to prevent update method from writing while sending
        data_to_send = self._get_data()
        if any(data_to_send.values()):
            errored_channels = []
            succeeded_channels= []

            # Send to channels
            for channel in self.channels:
                # Clear previous messages sent to channel if mode is MODE_DELETE_SEND
                if core.GLOBALS.is_user:
                    await asyncio.sleep(RLIM_USER_WAIT_TIME)
                context = await self._send_channel(channel, **data_to_send)
                if context["success"]:
                    succeeded_channels.append(channel)
                else:
                    errored_channels.append({"channel":channel, "reason": context["reason"]})

            # Remove any channels that returned with code status 404 (They no longer exist)
            for data in errored_channels:
                reason = data["reason"]
                channel = data["channel"]
                if isinstance(reason, discord.HTTPException):
                    if (reason.status == 403 or                    # Forbidden
                        reason.code in {50007, 10003}     # Not Forbidden, but bad error codes
                    ):
                        self.channels.remove(channel)
                        trace(f"Channel {channel.name}(ID: {channel.id}) {'was deleted' if reason.code == 10003 else 'does not have permissions'}, removing it from the send list", TraceLEVELS.WARNING)

            # Return sent data + failed and successful function for logging purposes
            return self._generate_log_context(**data_to_send, succeeded_ch=succeeded_channels, failed_ch=errored_channels)

        return None

    @misc._async_safe("update_semaphore")
    async def update(self, **kwargs: Any):
        """
        .. versionadded:: v2.0

        Used for changing the initialization parameters the object was initialized with.
        
        .. warning::
            Upon updating, the internal state of objects get's reset, meaning you basically have a brand new created object.
        
        Parameters
        -------------
        **kwargs: Any
            Custom number of keyword parameters which you want to update, these can be anything that is available during the object creation.
        
        Raises
        -----------
        DAFParameterError(code=DAF_UPDATE_PARAMETER_ERROR)
            Invalid keyword argument was passed
        Other
            Raised from .initialize() method.
        """
        if "start_now" not in kwargs:
            # This parameter does not appear as attribute, manual setting necessary
            kwargs["start_now"] = True
        
        await core._update(self, **kwargs) # No additional modifications are required
 
@sql._register_type("MessageTYPE")
class DirectMESSAGE(BaseMESSAGE):
    """
    This class is used for creating objects that represent messages which will be sent to Discord's TEXT CHANNELS.

    .. versionchanged:: v2.0
        
        - Channels parameter now also accepts channel objects instead of int.
        - ``.update`` method added.

    Parameters
    ------------
    start_period: Union[int, None]
        The value of this parameter can be:

        ..  table:: 
        
            ===========  =================================================================================================================
             Value        Info
            ===========  =================================================================================================================
             None         Messages are sent in a constant time period equal to the value of ``end_period``.
             int > 0      Messages are sent in a randomized time period. ``start_period`` represents the bottom limit of this period.
            ===========  =================================================================================================================

    end_period: int
        If ``start_period`` > 0, then this represents the upper limit of randomized time period in which messages will be sent.
        If ``start_period`` is None, then this represents the actual time period between each message send.

        .. code-block:: python
            :caption: **Randomized** sending period between **5** seconds and **10** seconds.
            
            # Time between each send is somewhere between 5 seconds and 10 seconds.
            framework.DirectMESSAGE(start_period=5, end_period=10, data="Second Message",  mode="send", start_now=True)

        .. code-block:: python
            :caption: **Fixed** sending period at **10** seconds

            # Time between each send is exactly 10 seconds.
            framework.DirectMESSAGE(start_period=None, end_period=10, data="Second Message",  mode="send", start_now=True)

    data: Union[str, EMBED, FILE, List[Union[str, EMBED, FILE]], _FunctionBaseCLASS]
        The data parameter is the actual data that will be sent using discord's API. The data types of this parameter can be:
            - str (normal text),
            - :ref:`EMBED`,
            - :ref:`FILE`,
            - List/Tuple containing any of the above arguments (There can up to 1 string, up to 1 :ref:`EMBED` and up to 10 :ref:`FILE` objects. If more than 1 string or embeds are sent, the framework will only consider the last found).
            - Function that accepts any amount of parameters and returns any of the above types. To pass a function, YOU MUST USE THE :ref:`data_function` decorator on the function before passing the function to the framework.

    mode: str
        Parameter that defines how message will be sent to a channel.
        It can be:   

        .. table:: 
            :align: left

            =================  =======================================================================================
              Mode               Description
            =================  =======================================================================================
             "send"             each period a new message will be sent,                                               
             "edit"             each period the previously send message will be edited (if it exists)                 
             "clear-send"       previous message will be deleted and a new one sent.                                  
            =================  =======================================================================================

    start_now: bool
        If True, then the framework will send the message as soon as it is run.
    """

    __slots__ = (
        "mode",
        "previous_message",
        "dm_channel",
    )
    __logname__ = "DirectMESSAGE"               # Used for logging (type key) and sql lookup table type registration
    __valid_data_types__ = {str, EMBED, FILE}   # Defines the allowed data types for the data parameter (get's checked in the ._initialize_data method)                                             

    def __init__(self,
                 start_period: Union[float, None],
                 end_period: float,
                 data: Union[str, EMBED, FILE, list],
                 mode: Literal["send", "edit", "clear-send"] = "send",
                 start_now: bool = True):
        super().__init__(start_period, end_period, data, start_now)
        self.mode = mode
        self.dm_channel = None
        self.previous_message = None

    def _generate_log_context(self,
                              success_context: Dict[bool, Exception],
                              text : str,
                              embed : EMBED,
                              files : List[FILE]):
        """
        Generates information about the message send attempt that is to be saved into a log.

        Parameters
        -----------
        text: str
            The text that was sent.
        embed: EMBED
            The embed that was sent.
        files: List[FILE]
            List of files that were sent.
        success_context: Dict[bool, Exception]
            Dictionary containing information about succession of the DM attempt. 
            Contains "success": `bool` key and "reason": `Exception` key which is only present if "success" is `False`


        Returns
        ----------
        Dict[str, Any]
            Dictionary containing:

                - sent_data: Dict[str, str]:
                    - streamed_audio - The filename that was streamed/youtube url

                - success_info: Dict[str, Any]:
                    - success: bool - Was sending successful or not
                    - reason:  str  - If it was unsuccessful, what was the reason
                    - delete: bool  - Signals the guild object to remove this message from the list
                                      due to unrecoverable error (if set to True).
                
                - type: str - The type of the message, this is always TextMESSAGE.
                - mode: str - The mode used to send the message (send, edit, clear-send).
        """
        embed = embed.to_dict() if embed is not None else None
        files = [x.filename for x in files]
        if not success_context["success"]:
            success_context["reason"] = str(success_context["reason"])

        sent_data_context = {}
        if text is not None:
            sent_data_context["text"] = text
        if embed is not None:
            sent_data_context["embed"] = embed
        if len(files):
            sent_data_context["files"] = files
        return {
            "sent_data": {
                **sent_data_context
            },
            "success_info": {
                **success_context
            },
            "type" : type(self).__name__,
            "mode" : self.mode
        }

    def _get_data(self) -> dict:
        """
        Returns a dictionary of keyword arguments that is then expanded
        into other functions (_send_channel, generate_log).
        This is exactly the same as in :ref:`TextMESSAGE`
        """
        return TextMESSAGE._get_data(self)

    async def _initialize_channels(self,
                                  user: discord.User):
        """
        The method creates a direct message channel and
        returns True on success or False on failure

        Parameters
        -----------
        - user: discord.User - discord User object to whom the DM will be created for
        
        Raises
        ---------
        - DAFNotFoundError(code=DAF_USER_CREATE_DM) - Raised when the direct message channel could not be created
        """
        try:
            await user.create_dm()
            self.dm_channel = user
        except discord.HTTPException as ex:
            raise DAFNotFoundError(f"Unable to create DM with user {user.display_name}\nReason: {ex}", DAF_USER_CREATE_DM)

    async def _handle_error(self, ex: Exception) -> bool:
        """
        This method handles the error that occurred during the execution of the function.
        Returns `True` if error was handled.

        Parameters
        -----------
        ex: Exception
            The exception that occurred during a send attempt.
        """
        handled = False
        if isinstance(ex, discord.HTTPException):
            if ex.status == 429 or ex.code == 40003: # Too Many Requests or opening DMs too fast
                retry_after = float(ex.response.headers["Retry-After"])  * RLIM_SAFETY_FACTOR
                trace(f"Rate limited, sleeping for {retry_after} seconds", TraceLEVELS.WARNING)
                await asyncio.sleep(retry_after)
                handled = True
            elif ex.status == 404:      # Unknown object
                if ex.code == 10008:    # Unknown message
                    self.previous_message  = None
                    handled = True
            elif ex.status == 403 or ex.code in {50007, 10001, 10003}:
                self._delete()

            if ex.status in {400, 403}: # Bad Request
                await asyncio.sleep(RLIM_USER_WAIT_TIME * 5) # To avoid triggering selfbot detection
                        
        return handled

    async def _send_channel(self,
                           text: str,
                           embed: EMBED,
                           files: List[FILE]) -> dict:
        """
        Sends data to the DM channel (user).
        
        Returns
        ------------
        Returns a dictionary:
        - "success" - Returns True if successful, else False.
        - "reason"  - Only present if "success" is False, contains the Exception returned by the send attempt.
        """
        # Send/Edit messages
        for tries in range(3):  # Maximum 3 tries (if rate limit)
            try:
                # Deletes previous message if it exists and mode is "clear-send"
                if self.mode == "clear-send" and self.previous_message is not None:
                    await self.previous_message.delete()
                    self.previous_message = None

                # Sends a new message
                if (self.mode in {"send" , "clear-send"} or
                     self.mode == "edit" and self.previous_message is None
                ):
                    self.previous_message = await self.dm_channel.send(text, embed=embed,
                                                                       files=[discord.File(fwFILE.filename) for fwFILE in files])

                # Mode is edit and message was already send to this channel
                elif self.mode == "edit":
                    await self.previous_message.edit(text, embed=embed)
                return {"success" : True}

            except Exception as ex:
                if await self._handle_error(ex) is False or tries == 2:
                    return {"success" : False, "reason" : ex}

    @misc._async_safe("update_semaphore")
    async def send(self) -> Union[dict, None]:
        """
        Sends the data into the channels
        
        Returns
        ----------
        Union[Dict, None]
            Returns a dictionary generated by the ``_generate_log_context`` method or the None object if message wasn't ready to be sent (:ref:`data_function` returned None or an invalid type)
            
            This is then passed to :ref:`GUILD`.generate_log method.
        """
        # Parse data from the data parameter
        data_to_send = self._get_data()
        if any(data_to_send.values()):
            if core.GLOBALS.is_user:
                await asyncio.sleep(RLIM_USER_WAIT_TIME)
            context = await self._send_channel(**data_to_send)

            return self._generate_log_context(context, **data_to_send)

        return None

    @misc._async_safe("update_semaphore")
    async def update(self, init_options={},**kwargs):
        """
        .. versionadded:: v2.0

        Used for changing the initialization parameters the object was initialized with.
        
        .. warning:: 
            Upon updating, the internal state of objects get's reset, meaning you basically have a brand new created object.
        
        Parameters
        -------------
        init_options: Dict
            Contains additional initialization options, not meant for public use, should only be called by the USER update method recursively.
        **kwargs: Any
            Custom number of keyword parameters which you want to update, these can be anything that is available during the object creation.

        Raises
        -----------
        DAFParameterError(code=DAF_UPDATE_PARAMETER_ERROR)
            Invalid keyword argument was passed
        Other
            Raised from .initialize() method
        """

        if "start_now" not in kwargs:
            # This parameter does not appear as attribute, manual setting necessary
            kwargs["start_now"] = True

        if not len(init_options):
            init_options = {"user" : self.dm_channel}
        
        await core._update(self, init_options=init_options, **kwargs) # No additional modifications are required



"""
    This modules contains definitions related to the client (for API)
"""

import  asyncio
import  _discord as discord
from    .tracing import *
from    . import core


#######################################################################
# Globals
#######################################################################
class GLOBALS:
    client = None     # Pycord Client object
    running = False # Weird bug causes client.on_ready method to be called multiple times some times

__all__ = (
    "CLIENT",
    "get_client"
)


class CLIENT(discord.Client):
    """
    The same as `discord.Client <https://docs.pycord.dev/en/master/api.html?highlight=client#discord.Client>`_, except it contains an on_ready coroutine.

    .. note::
        This is automatically created by the framework.
        You can retrieve the object created by calling :ref:`get_client` function.
    """
    async def on_ready(self) -> None:
        """
        Gets run at login, creates the main initialization task inside the core module.
        """
        if GLOBALS.running:
            return

        GLOBALS.running = True
        trace(f"[CLIENT]: Logged in as {self.user}", TraceLEVELS.NORMAL)
        # Initialize all the modules from the core module
        self.loop.create_task(core._initialize())
        

def _initialize(token: str, *,
               bot: bool,
               intents: discord.Intents):
    """
    
    Initializes the client module (Pycord).

    Parameters
    ----------------
    token: str
        Authorization token for connecting to discord.
    bot: bool
        Tells if the token is for a bot account.
                      and False if the token is  for an user account.
    intents: discord.Intents
        The intents discord object. Intents are settings that
                                    dictate which dictates the events that the client will listen for.
    """
    GLOBALS.client = CLIENT(intents=intents)
    if not bot:
        trace("[CLIENT]: Bot is an user account which is against discord's ToS",TraceLEVELS.WARNING)
    GLOBALS.client.run(token, bot=bot)


def get_client() -> CLIENT:
    """
    Returns the `CLIENT` object used for communicating with Discord.
    """
    return GLOBALS.client



"""
    Contains definitions related to voice messaging."""

from   .base        import *
from   ..dtypes     import *
from   ..tracing    import *
from   ..const      import *
from   ..exceptions import *
from   typing       import Any, Dict, List, Iterable, Union
from   ..           import client
from   ..           import sql
from   ..           import core
from   ..           import misc
import asyncio
import _discord as discord


__all__ = (
    "VoiceMESSAGE",
)

class GLOBALS:
    """ ~ class ~
    - @Info: Contains global variables used in the voice messaging.
    """
    voice_client: discord.VoiceClient = None

@sql._register_type("MessageTYPE")
class VoiceMESSAGE(BaseMESSAGE):
    """
    This class is used for creating objects that represent messages which will be streamed to voice channels.

    .. versionchanged:: v2.0
        
        - Added the ``volume`` parameter
        - Renamed ``channel_ids`` parameter to ``channels``
        - Channels parameter now also accepts channel objects instead of int

    Parameters
    ------------
    start_period: Union[int, None]
        The value of this parameter can be:

        ..  table:: 
        
            ===========  =================================================================================================================
             Value        Info
            ===========  =================================================================================================================
             None         Messages are sent in a constant time period equal to the value of ``end_period``.
             int > 0      Messages are sent in a randomized time period. ``start_period`` represents the bottom limit of this period.
            ===========  =================================================================================================================

    end_period: int
        If ``start_period`` > 0, then this represents the upper limit of randomized time period in which messages will be sent.
        If ``start_period`` is None, then this represents the actual time period between each message send.

        .. code-block:: python
            :caption: **Randomized** sending period between **5** seconds and **10** seconds.
            
            # Time between each send is somewhere between 5 seconds and 10 seconds.
            framework.VoiceMESSAGE(start_period=None, end_period=10, data=framework.AUDIO("msg.mp3"), channels=[12345], start_now=True, volume=50)

        .. code-block:: python
            :caption: **Fixed** sending period at **10** seconds

            # Time between each send is exactly 10 seconds.
            framework.VoiceMESSAGE(start_period=None, end_period=10, data=framework.AUDIO("msg.mp3"), channels=[12345], start_now=True, volume=50)

    data: AUDIO
        The data parameter is the actual data that will be sent using discord's API. The data types of this parameter can be:
            - AUDIO object.
            - Function that accepts any amount of parameters and returns an AUDIO object. To pass a function, YOU MUST USE THE :ref:`data_function` decorator on the function before passing the function to the framework.
    channels: Iterable[Union[int, discord.VoiceChannel]]
        Channels that it will be advertised into (Can be snowflake ID or channel objects from PyCord).
    volume: int         
        The volume (0-100%) at which to play the audio. Defaults to 50%. This was added in v2.0.0
    start_now: bool
        If True, then the framework will send the message as soon as it is run.      
    """

    __slots__ = (
        "randomized_time",
        "period",
        "start_period",
        "end_period",
        "data",
        "volume",
        "channels",
        "timer",
        "force_retry",
    )

    __logname__ = "VoiceMESSAGE"    # For sql._register_type
    __valid_data_types__ = {AUDIO}  # This is used in the BaseMESSAGE.initialize() to check if the passed data parameters are of correct type

    def __init__(self, start_period: Union[float, None],
                 end_period: float,
                 data: AUDIO,
                 channels: Iterable[Union[int, discord.VoiceChannel]],
                 volume: int=50,
                 start_now: bool = True):

        super().__init__(start_period, end_period, data, start_now)
        self.volume = max(0, min(100, volume)) # Clamp the volume to 0-100 % 
        self.channels = list(set(channels))    # Auto remove duplicates

    def _generate_log_context(self,
                             audio: AUDIO,
                             succeeded_ch: List[discord.VoiceChannel],
                             failed_ch: List[dict]) -> Dict[str, Any]:
        """
        Generates information about the message send attempt that is to be saved into a log.

        Parameters
        -----------
        audio: audio
            The audio that was streamed.
        succeeded_ch: List[Union[discord.VoiceChannel]] 
            List of the successfully streamed channels
        failed_ch: List[Dict[discord.VoiceChannel, Exception]] 
            List of dictionaries contained the failed channel and the Exception object

        Returns
        ----------
        Dict[str, Any]
            Dictionary containing:

                - sent_data: Dict[str, str]:
                    - streamed_audio: str - The filename that was streamed/youtube url

                - channels: Dict[str, List]:
                    - successful: List[Dict[str, int]] - List of dictionaries containing name of the channel and snowflake id of the channels.
                    - failed: List[Dict[str, Any]] - List of dictionaries containing name of the channel (str), snowflake id (int) and reason why streaming to channel failed (str).
                
                - type: str - The type of the message, this is always VoiceMESSAGE.
        """

        succeeded_ch = [{"name": str(channel), "id" : channel.id} for channel in succeeded_ch]
        failed_ch = [{"name": str(entry["channel"]), "id" : entry["channel"].id,
                     "reason": str(entry["reason"])} for entry in failed_ch]
        return {
            "sent_data": {
                "streamed_audio" : audio.to_dict()
            },
            "channels": {
                "successful" : succeeded_ch,
                "failed": failed_ch
            },
            "type" : type(self).__name__
        }

    def _get_data(self) -> dict:
        """"
        Returns a dictionary of keyword arguments that is then expanded
        into other methods eg. `_send_channel, generate_log`
        """
        data = None
        _data_to_send = {}
        data = self.data.get_data() if isinstance(self.data, _FunctionBaseCLASS) else self.data
        if data is not None:
            if not isinstance(data, (list, tuple, set)):
                data = (data,)
            for element in data:
                if isinstance(element, AUDIO):
                    _data_to_send["audio"] = element
        return _data_to_send

    async def _initialize_channels(self):
        """
        This method initializes the implementation specific api objects and checks for the correct channel input context.
        
        Raises
        ------------
        - `DAFParameterError(code=DAF_INVALID_TYPE)` - Raised when the object retrieved from channels is not a discord.TextChannel or discord.Thread object.
        - `DAFNotFoundError(code=DAF_MISSING_PARAMETER)` - Raised when no channels could be found were parsed.
        """
        ch_i = 0
        cl = client.get_client()
        while ch_i < len(self.channels):
            channel = self.channels[ch_i]
            if isinstance(channel, discord.abc.GuildChannel):
                channel_id = channel.id
            else:
                channel_id = channel
                channel = self.channels[ch_i] = cl.get_channel(channel_id)

            if channel is None:
                trace(f"Unable to get channel from ID {channel_id}", TraceLEVELS.ERROR)
                self.channels.remove(channel)
            elif type(channel) not in {discord.VoiceChannel}:
                raise DAFParameterError(f"TextMESSAGE object received channel type of {type(channel).__name__}, but was expecting VoiceChannel", DAF_INVALID_TYPE)
            else:
                ch_i += 1

        if not len(self.channels):
            raise DAFNotFoundError(f"No valid channels were passed to {type(self)} object", DAF_MISSING_PARAMETER)

    async def _send_channel(self,
                           channel: discord.VoiceChannel,
                           audio: AUDIO) -> dict:
        """
        Sends data to specific channel
        
        Returns a dictionary:
        - "success" - Returns True if successful, else False
        - "reason"  - Only present if "success" is False, contains the Exception returned by the send attempt
        
        Parameters
        -------------
        channel: discord.VoiceChannel
            The channel in which to send the data.
        audio: AUDIO
            the audio to stream.
        """
        stream = None
        try:
            # Check if client has permissions before attempting to join
            ch_perms = channel.permissions_for(channel.guild.get_member(client.get_client().user.id))
            if not all([ch_perms.connect, ch_perms.stream, ch_perms.speak]):
                raise self._generate_exception(403, 50013, "You lack permissions to perform that action", discord.Forbidden)

            if GLOBALS.voice_client is None or not GLOBALS.voice_client.is_connected():
                GLOBALS.voice_client = await channel.connect(timeout=C_VC_CONNECT_TIMEOUT)
            else:
                await GLOBALS.voice_client.move_to(channel)

            stream = discord.PCMVolumeTransformer(discord.FFmpegPCMAudio(audio.url), volume=self.volume/100)

            GLOBALS.voice_client.play(stream)

            while GLOBALS.voice_client.is_playing():
                await asyncio.sleep(1)
            return {"success": True}
        except Exception as ex:
            if isinstance(ex, (FileExistsError, discord.Forbidden)):
                pass # Don't change error
            elif client.get_client().get_channel(channel.id) is None:
                ex = self._generate_exception(404, 10003, "Channel was deleted", discord.NotFound)
            else:
                ex = self._generate_exception(500, 0, "Timeout error", discord.HTTPException)
            return {"success": False, "reason": ex}
        finally:
            if GLOBALS.voice_client is not None and GLOBALS.voice_client.is_connected():
                await GLOBALS.voice_client.disconnect()
                GLOBALS.voice_client = None
                await asyncio.sleep(1) # Avoid sudden disconnect and connect to a new channel

    @misc._async_safe("update_semaphore")
    async def send(self) -> Union[dict,  None]:
        """
        Sends the data into each channel.
        
        Returns
        ----------
        Union[Dict, None]
            Returns a dictionary generated by the ``_generate_log_context`` method or the None object if message wasn't ready to be sent (:ref:`data_function` returned None or an invalid type)
            
            This is then passed to :ref:`GUILD`.generate_log method.
        """
        _data_to_send = self._get_data()
        if any(_data_to_send.values()):
            errored_channels = []
            succeeded_channels= []

            for channel in self.channels:
                context = await self._send_channel(channel, **_data_to_send)
                if context["success"]:
                    succeeded_channels.append(channel)
                else:
                    errored_channels.append({"channel":channel, "reason": context["reason"]})

            # Remove any channels that returned with code status 404 (They no longer exist)
            for data in errored_channels:
                reason = data["reason"]
                channel = data["channel"]
                if isinstance(reason, discord.HTTPException):
                    if (reason.status == 403 or
                        reason.code in {10003, 50013} # Unknown, Permissions
                    ):
                        self.channels.remove(channel)
                        trace(f"Channel {channel.name}(ID: {channel.id}) {'was deleted' if reason.code == 10003 else 'does not have permissions'}, removing it from the send list", TraceLEVELS.WARNING)

            return self._generate_log_context(**_data_to_send, succeeded_ch=succeeded_channels, failed_ch=errored_channels)

        return None

    @misc._async_safe("update_semaphore")
    async def update(self, **kwargs):
        """
        .. versionadded:: v2.0

        Used for changing the initialization parameters the object was initialized with.
        
        .. warning::
            Upon updating, the internal state of objects get's reset, meaning you basically have a brand new created object.
        
        Parameters
        -------------
        **kwargs: Any
            Custom number of keyword parameters which you want to update, these can be anything that is available during the object creation.
        
        Raises
        -----------
        DAFParameterError(code=DAF_UPDATE_PARAMETER_ERROR)
            Invalid keyword argument was passed
        Other
            Raised from .initialize() method
        """
        if "start_now" not in kwargs:
            # This parameter does not appear as attribute, manual setting necessary 
            kwargs["start_now"] = True

        await core._update(self, **kwargs) # No additional modifications are required



"""
Contains the definitions related to errors that can be raised inside the framework.
"""

class DAFError(Exception):
    """
    Base exception class for all DAF exceptions.
    
    Parameters
    -------------
    message: str
        The exception message.
    code: int
        The error code.
    """
    def __init__(self, message: str, code: int):
        self.code = code # Error code
        super().__init__(message)

class DAFParameterError(DAFError):
    """
    Raised when theres an parameter exception.
    
    Parameters
    -------------
    message: str
        The exception message.
    code: int
        The error code.
    """
    pass

class DAFNotFoundError(DAFError):
    """
    Raised when an object is not found.

    Parameters
    -------------
    message: str
        The exception message.
    code: int
        The error code.
    """
    pass

class DAFSQLError(DAFError):
    """
    Raised whenever there's an error with SQL.

    Parameters
    -------------
    message: str
        The exception message.
    code: int
        The error code.
    """
    pass

# Error Codes
# Guild codes
DAF_GUILD_ALREADY_ADDED         = 0     #: Guild with specified snowflake is already added.
DAF_GUILD_ID_REQUIRED           = 1     #: Guild ID is required but was not passed.    
DAF_GUILD_ID_NOT_FOUND          = 2     #: Guild with specified snowflake was not found (or user).
DAF_USER_CREATE_DM              = 3     #: Was unable to create DM with user (probably user not found).
# Data type codes
DAF_INVALID_TYPE                = 4     #: Object of invalid type was given.
DAF_YOUTUBE_STREAM_ERROR        = 5     #: The given youtube link could not be streamed (AUDIO, VoiceMESSAGE).
DAF_FILE_NOT_FOUND              = 6     #: The given file was not found.
# Parameter specific errors
DAF_UPDATE_PARAMETER_ERROR      = 7     #: The update method only accepts the following keyword arguments.
DAF_MISSING_PARAMETER           = 8     #: The parameter(s) is(are) missing.
# SQL Exceptions
DAF_SQL_CREATE_TABLES_ERROR     = 9     #: Unable to create all the tables.
DAF_SQL_LOOKUPTABLE_NOT_FOUND   = 10    #: The lookup table was not found.
DAF_SQL_BEGIN_ENGINE_ERROR      = 11    #: Unable to start engine.
DAF_SQL_CR_LT_VALUES_ERROR      = 12    #: Unable to create lookuptables' rows.
DAF_SQL_CREATE_DT_ERROR         = 13    #: Unable to create SQL data types.
DAF_SQL_CREATE_VPF_ERROR        = 14    #: Unable to create views, procedures and functions.
DAF_SQL_CURSOR_CONN_ERROR       = 15    #: Unable to connect the cursor.



"""
    This module contains definitions regarding miscellaneous
    items that can appear in multiple modules
"""

from typing import Coroutine, Callable, Any, Optional
from asyncio import Semaphore


###############################
# Safe access functions
###############################
def _write_attr_once(obj: Any, name: str, value: Any):
    """
    Method that assigns an attribute only if it does not already have a reference to an object.
    This is to prevent any ``.update()`` method calls from resetting critical variables that should not be changed,
    even if the objects goes thru initialization again.

    Parameters
    -------------
    obj: Any
        Object to safely write.
    
    name: str
        The name of the attribute to change.

    value: Any
        The value to change the attribute with.
    """

    # TODO: Test if it works
    if not hasattr(obj, name): # Write only if forced, or if not forced, then the attribute must not exist
        setattr(obj, name, value)



###########################
# Decorators
###########################
# TODO: Change Lock to semaphore, add parameter that dictates how many to take
def _async_safe(semaphore: str, amount: Optional[int]=1) -> Callable:
    """
    Function that returns a safety decorator, which uses the :strong:`semaphore` parameter
    as a safety mechanism.

    This is for usage on :strong:`methods`

    Parameters
    ----------------
    semaphore: str
        Name of the semaphore attribute to take.
    amount: Optional[int]
        How many times to take the semaphore.

    Returns
    --------------
    Returns the safety decorator.

    Raises
    --------------
    TypeError
        The ``semaphore`` parameter is not a string describing the semaphore attribute of a table.
    """

    def __safe_access(coroutine: Coroutine) -> Coroutine:
        """
        Decorator that returns a method wrapper Coroutine that utilizes a
        asyncio semaphore to assure safe asynchronous operations.
        """
        async def wrapper(self, *args, **kwargs):
            sem: Semaphore = getattr(self, semaphore)
            for i in range(amount):
                await sem.acquire()

            result = await coroutine(self, *args, **kwargs)

            for i in range(amount):
                sem.release()

            return result
        
        wrapper.__annotations__ = coroutine.__annotations__ # Keep the same type hints

        return wrapper


    if type(semaphore) is not str:
        raise TypeError("semaphore parameter must be an attribute name of the asyncio semaphore inside the object")

    return __safe_access


"""
    This module contains the essential definitions
    and functions needed for the framework to run,
    as well as user function to control the framework
"""
from   typing import Any, Dict, Iterable, Literal, Callable, List, Optional, Union, overload
import asyncio
import copy
import inspect
import _discord as dc
from . const import *
from . exceptions import *
from . import tracing
from . tracing import *
from . import guild
from . import client
from . import sql
from . import message


#######################################################################
# Exports
#######################################################################
__all__ = (
    "run",
    "shutdown",
    "add_object",
    "remove_object",
)

#######################################################################
# Globals   (These are all set in the framework.run function)
#######################################################################
class GLOBALS:
    """
    Storage class used for holding global variables.
    """
    user_callback: Callable = None
    server_list: List[guild._BaseGUILD] = []
    temp_server_list: List[guild._BaseGUILD] = None # Holds the guilds that are awaiting initialization (set in framework.run and cleared after initialization)
    sql_manager: sql.LoggerSQL = None,
    is_user: bool = False


#######################################################################
# Tasks
#######################################################################
async def _advertiser(message_type: Literal["text", "voice"]) -> None:
    """
    The task that is responsible for shilling to channels.
    This is the most top level task.
    
    Parameters
    ------------
    message_type: str
        Two tasks advertising tasks are created, this variable tells the guild objects which
        task is requesting to shill, so it knows what type of messages to actually send.
    """
    while True:
        await asyncio.sleep(C_TASK_SLEEP_DELAY)
        for guild_user in GLOBALS.server_list[:]: # Shallow copy list to allow dynamic removal and addition of objects
            await guild_user.advertise(message_type)


#######################################################################
# Functions
#######################################################################
async def _initialize() -> None:
    """
    The main initialization function.
    It initializes all the other modules, creates advertising tasks
    and initializes all the core functionality.
    """
    # Initialize the SQL module if manager is provided
    # If manager is not provided, use JSON based file logs
    sql_manager = GLOBALS.sql_manager
    _client = client.get_client()
    if sql_manager is not None:
        try:
            await sql.initialize(sql_manager) # Initialize the SQL database
        except DAFSQLError as ex:
            trace(f"Unable to initialize the SQL manager, JSON logs will be used.\nReason: {ex}", TraceLEVELS.WARNING)
    else:
        trace("[CORE]: No SQL manager provided, logging will be JSON based", TraceLEVELS.NORMAL)

    # Initialize the servers (and their message objects)
    trace("[CORE]: Initializing servers", TraceLEVELS.NORMAL)
    for server in GLOBALS.temp_server_list:
        try:
            await add_object(server) # Add each guild to the shilling list
        except DAFError as ex:
            trace(ex)

    # Create advertiser tasks
    trace("[CORE]: Creating advertiser tasks", TraceLEVELS.NORMAL)
    _client.loop.create_task(_advertiser("text"))
    _client.loop.create_task(_advertiser("voice"))

    # Create the user callback task
    callback = get_user_callback()
    if callback is not None:
        trace("[CORE]: Starting user callback function", TraceLEVELS.NORMAL)
        _client.loop.create_task(callback)

    del GLOBALS.sql_manager         # Variable is no longer needed, instead the sql_manager inside sql.py is used
    del GLOBALS.temp_server_list    # Variable is no longer needed

    trace("[CORE]: Initialization complete.", TraceLEVELS.NORMAL)


@overload
async def add_object(obj: Union[guild.USER, guild.GUILD]) -> None: 
    """

    Adds a guild or an user to the framework.
    
    Parameters
    --------------
    obj: Union[guild.USER, guild.GUILD]
        The guild object to add into the framework.

    Raises
    -----------
    DAFParameterError(code=DAF_GUILD_ALREADY_ADDED)
         The guild/user is already added to the framework.
    DAFParameterError(code=DAF_INVALID_TYPE)
         The object provided is not supported for addition.
    Other
        Raised in the obj.initialize() method
    """
    ...
@overload
async def add_object(obj: Union[message.DirectMESSAGE, message.TextMESSAGE, message.VoiceMESSAGE],
                     snowflake: Union[int, guild.GUILD, guild.USER, dc.Guild, dc.User]) -> None:
    """

    Adds a message to the framework.
    
    Parameters
    --------------
    obj: Union[message.DirectMESSAGE, message.TextMESSAGE, message.VoiceMESSAGE]
        The message object to add into the framework.
    snowflake: Union[int, guild.GUILD, guild.USER, dc.Guild, dc.User]
        Which guild/user to add it to (can be snowflake id or a framework _BaseGUILD object or a discord API wrapper object).

    Raises
    -----------
    DAFParameterError(code=DAF_GUILD_ID_REQUIRED)
         guild_id wasn't provided when adding a message object (to which guild should it add)
    DAFNotFoundError(code=DAF_GUILD_ID_NOT_FOUND)
        Could not find guild with that id.
    DAFParameterError(code=DAF_INVALID_TYPE)
         The object provided is not supported for addition.
    Other
        Raised in the obj.add_message() method
    """
    ...
async def add_object(obj, snowflake=None):
    object_type_name = type(obj).__name__
    # Convert the `snowflake` object into a discord snowflake ID (only if adding a message to guild)
    if isinstance(snowflake, (dc.Guild, dc.User)):
        snowflake = snowflake.id
    elif isinstance(snowflake, guild._BaseGUILD):
        snowflake = snowflake.snowflake

    # Add the object
    if isinstance(obj, guild._BaseGUILD):
        if obj in GLOBALS.server_list:
            raise DAFParameterError(f"{object_type_name} with snowflake `{obj.snowflake}` is already added to the framework.", DAF_GUILD_ALREADY_ADDED)

        await obj.initialize()
        GLOBALS.server_list.append(obj)

    elif isinstance(obj, message.BaseMESSAGE):
        if snowflake is None:
            raise DAFParameterError(f"`snowflake` is required to add a message. Only the {object_type_name} object was provided.", DAF_GUILD_ID_REQUIRED)

        for guild_user in GLOBALS.server_list:
            if guild_user.snowflake == snowflake:
                await guild_user.add_message(obj)
                return

        raise DAFNotFoundError(f"Guild or user with snowflake `{snowflake}` was not found in the framework.", DAF_GUILD_ID_NOT_FOUND)

    else:
        raise DAFParameterError(f"Invalid object type `{object_type_name}`.", DAF_INVALID_TYPE)


@overload
def remove_object(guild_id: int) -> None: 
    """
    Removes a guild from the framework that has the given guild_id.
    
    Parameters
    -------------
    - guild_id: `int` - ID of the guild to remove.
    
    Raises
    --------------
    DAFNotFoundError(code=DAF_GUILD_ID_NOT_FOUND)
         Could not find guild with that id.
    DAFParameterError(code=DAF_INVALID_TYPE)
         The object provided is not supported for removal."""
    ...
@overload
def remove_object(channel_ids: Iterable[int]) -> None:
    """
    Removes messages that contain all the given channel ids.
    
    Parameters
    --------------
    channel_ids: Iterable
        The channel IDs that the message must have to be removed (it must have all of these).
    
    Raises
    --------------------
    DAFParameterError(code=DAF_INVALID_TYPE)
        The object provided is not supported for removal.
    """
    ...
def remove_object(data):    
    if isinstance(data, int): # Guild id
        for guild_user in GLOBALS.server_list:
            if guild_user.snowflake == data:
                GLOBALS.server_list.remove(guild_user)
                return
        raise DAFNotFoundError(f"Guild with snowflake `{data}` was not found in the framework.", DAF_GUILD_ID_NOT_FOUND)

    elif isinstance(data, Iterable): # Channel ids
        for guild_user in GLOBALS.server_list:
            if isinstance(guild_user, guild.GUILD): # USER doesn't have channels
                for message in guild_user.t_messages + guild_user.vc_messages:
                    msg_channels = [x.id for x in message.channels] # Generate snowflake list
                    if all(x in msg_channels for x in data): # If any of the channels in the set are in the message channel list
                        guild_user.remove_message(message)
    else:
        raise DAFParameterError(f"Invalid parameter type `{type(data)}`.", DAF_INVALID_TYPE)


async def _update(obj: Any, *, init_options: dict = {}, **kwargs):
        """
        .. versionadded:: v2.0

        Used for changing the initialization parameters the obj was initialized with.
        
        .. warning::
            Upon updating, the internal state of objects get's reset, meaning you basically have a brand new created object.   

        .. warning::
            This is not meant for manual use, but should be used only by the obj's method.

        Parameters
        -------------
        obj: Any
            The object that contains a .update() method.
        init_options: dict
            Contains the initialization options used in .initialize() method for re-initializing certain objects.
            This is implementation specific and not necessarily available.
        Other:
            Other allowed parameters are the initialization parameters first used on creation of the object.

        Raises
        ------------
        DAFParameterError(code=DAF_UPDATE_PARAMETER_ERROR)
            Invalid keyword argument was passed.
        Other
            Raised from .initialize() method.
        """
        init_keys = inspect.getfullargspec(obj.__init__).args # Retrieves list of call args
        init_keys.remove("self")
        current_state = copy.copy(obj) # Make a copy of the current object for restoration in case of update failure
        try:  
            for k in kwargs:
                if k not in init_keys:
                    raise DAFParameterError(f"Keyword argument `{k}` was passed which is not allowed. The update method only accepts the following keyword arguments: {init_keys}", DAF_UPDATE_PARAMETER_ERROR)
            # Most of the variables inside the object have the same names as in the __init__ function.
            # This section stores attributes, that are the same, into the `updated_params` dictionary and
            # then calls the __init__ method with the same parameters, with the exception of start_period, end_period and start_now parameters
            updated_params = {}
            for k in init_keys:
                # Store the attributes that match the __init__ parameters into `updated_params`
                updated_params[k] = kwargs[k] if k in kwargs else getattr(obj, k)

            # Call the implementation __init__ function and then initialize API related things
            obj.__init__(**updated_params)
            # Call additional initialization function (if it has one)
            if hasattr(obj, "initialize"):
                await obj.initialize(**init_options)
        
        except Exception:
            # In case of failure, restore to original attributes
            for k in type(obj).__slots__:
                setattr(obj, k, getattr(current_state, k))

            raise


async def shutdown() -> None:
    """
    Stops the framework and any user tasks.
    """
    cl = client.get_client()
    await cl.close()


def get_user_callback() -> Callable:
    """
    Returns the user coroutine.
    """
    return GLOBALS.user_callback


def run(token : str,
        server_list : Optional[List[Union[guild.GUILD, guild.USER]]]=[],
        is_user : Optional[bool] =False,
        user_callback : Optional[Callable]=None,
        server_log_output : Optional[str] ="History",
        sql_manager: Optional[sql.LoggerSQL]=None,
        intents: Optional[dc.Intents]=dc.Intents.default(),
        debug : Optional[bool]=True) -> None:
    """
    Runs the framework. 

    This is the very first thing that needs to be called to start the framework.
    
    Parameters
    ---------------
    token: str
        Discord's access token for account.
    server_list: Optional[List[Union[:ref:`GUILD`, :ref:`USER`]]
        Predefined server list (guild list) to shill.
    is_user: Optional[bool]
        Set to True if the token is for an user account.
    user_callback: Optional[Callable]
        Users coroutine (task) to run after the framework is run.
    server_log_output: Optional[str]
        Path where the server log files will be created.
    sql_manager: Optional[:ref:`LoggerSQL`]
        SQL manager object that will save logs into the database.
    intents: Optional[discord.Intents]
        Discord Intents object (represents settings to which events it will be listened to).
    debug: Optional[bool]
        Print trace message to the console, useful for debugging.
    """

    guild.GLOBALS.server_log_path = server_log_output               # Logging folder
    GLOBALS.temp_server_list = server_list                          # List of guild objects to iterate thru in the advertiser task
    GLOBALS.sql_manager = sql_manager                               # SQL manager object
    GLOBALS.is_user = is_user                                       # Is the token from an user account
    if user_callback is not None:
        GLOBALS.user_callback = user_callback()                     # Called after framework has started
    
    tracing.initialize(debug)                                       # Print trace messages to the console for debugging purposes
    client._initialize(token, bot=not is_user, intents=intents)
